"""
A collection of tools and EOTasks for image co-registration
"""

from .coregistration import ECCRegistrationTask, InterpolationType, PointBasedRegistrationTask, RegistrationTask

__version__ = "1.3.1"
